﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;

namespace HeuristicLab.Algorithms.Bandits.BanditPolicies {
  // also called softmax policy
  public class BoltzmannExplorationPolicy : IBanditPolicy {
    private readonly double beta;
    private readonly Func<DefaultPolicyActionInfo, double> valueFunction;

    public BoltzmannExplorationPolicy(double eps) : this(eps, DefaultPolicyActionInfo.AverageReward) { }

    public BoltzmannExplorationPolicy(double beta, Func<DefaultPolicyActionInfo, double> valueFunction) {
      if (beta < 0) throw new ArgumentException();
      this.beta = beta;
      this.valueFunction = valueFunction;
    }
    public int SelectAction(Random random, IEnumerable<IBanditPolicyActionInfo> actionInfos) {
      Debug.Assert(actionInfos.Any());

      // select best
      var myActionInfos = actionInfos.OfType<DefaultPolicyActionInfo>();
      Debug.Assert(myActionInfos.Any(a => !a.Disabled));

      var w = from aInfo in myActionInfos
              select aInfo.Disabled
                ? 0.0
                : Math.Exp(beta * valueFunction(aInfo));

      var bestAction = myActionInfos
        .Select((aInfo, idx) => new { aInfo, idx })
        .SampleProportional(random, w)
        .Select(p => p.idx)
        .First();
      Debug.Assert(bestAction >= 0);
      return bestAction;
    }

    public IBanditPolicyActionInfo CreateActionInfo() {
      return new DefaultPolicyActionInfo();
    }

    public override string ToString() {
      return string.Format("BoltzmannExplorationPolicy({0:F2})", beta);
    }
  }
}
